// https://stackoverflow.com/questions/41476063/typescript-remove-key-from-type-subtraction-type
declare namespace PChannel {
    type Omit<O, D extends (string|number|Symbol)> = Pick<O, Exclude<keyof O, D>>;
}

declare namespace Types { 
    type ErasedType = string&number;

    export type Any = Dictionary | Primitive;
    export interface Dictionary { [key: string]: Any }
    export type Primitive = string|number|boolean|null|undefined|void;
    export type PrimitiveAnd = string&number&boolean&null&undefined&void;


    //Eh... we need to hack around `type X = X[]` not being allowed, even though it is not really circular.
    //type AnyWithoutArray = Any[];
    type AnyArrNoObject = Dictionary | Primitive;
    export type AnyArr = AnyArrNoObject | object;
    
    export interface DictionaryArr { [key: string]: AnyArr }
    export type Arr = Array<AnyArr | AnyArr[] | AnyArr[][] | AnyArr[][][] | AnyArr[][][][] | AnyArr[][][][][]>
    export type AnyAll = AnyArr | Arr;


    type MakeArr<T> = T | T[] | T[][] | T[][][] | T[][][][] | T[][][][][];
    export type AnyAllNoObject = MakeArr<Primitive | { [key in string]: AnyAllNoObject }>;

    export type AnyAllNoObjectBuffer = MakeArr<Uint8Array | Primitive | { [key in string]: AnyAllNoObjectBuffer }>;// | { [key: number]: AnyAllNoObjectBuffer };

    
    type MakeAndArr<T> = T & T[] & T[][] & T[][][] & T[][][][] & T[][][][][];
    export type AnyAndAll = MakeAndArr<PrimitiveAnd & { [key in string]: AnyAndAll }>;



    export type Graph<T> = { value: T; children: {[key: string]: Graph<T>} };
    export type GraphParent<T> = {
        value: T;
        children: {[key: string]: GraphParent<T>};
        parent: GraphParent<T>|undefined;
        path: string[];
        pathHash: string;
    };

    export type DeepReadonly<T> = {
        readonly [P in keyof T]: DeepReadonly<T[P]>;
    }

    /*
    TODO: Use this when we use TS 2.8, (except use our own primitive)
    export type primitive = string | number | boolean | undefined | null
    export type DeepReadonly<T> =
        T extends primitive ? T :
        T extends Array<infer U> ? DeepReadonlyArray<U> :
        DeepReadonlyObject<T>
    ;

    export interface DeepReadonlyArray<T> extends ReadonlyArray<DeepReadonly<T>> {}

    export type DeepReadonlyObject<T> = {
    readonly [P in keyof T]: DeepReadonly<T[P]>
    }
    */
}

declare var NODE: boolean;
declare var TEST: boolean;

declare var PROMISE_defaultTimeout: number;

declare module "reflection/misc" {
    export let g: any;
    export function IsEmpty<T>(obj: {
        [key: string]: T;
    }): boolean;
    export function Range(start: number, end: number): number[];
}



declare module "controlFlow/pChan" {
    export function SetDefaultTimeout(timeout: number | undefined, code: () => void): void;
    /** An alias for new PChan. Lowercase so it is super short and easy to type, so I can use it everywhere. */
    export function pchan<T>(promiseErrorTimeout?: number): PChan<T>;
    export interface PChanReceive<T> {
        OnClosed: Promise<void>;
        HasValues(): boolean;
        GetPromise(): Promise<T>;
        IsClosed(): boolean;
        Close(): void;
        IsClosedError(err: any): boolean;
    }
    export interface PChanSend<T> {
        OnClosed: Promise<void>;
        SendValue(...values: T extends void ? [] | [T | PromiseLike<T>] : [T | PromiseLike<T>]): void;
        SendError(err: any): void;
        IsClosed(): boolean;
        Close(): void;
        IsClosedError(err: any): boolean;
    }
    export class PChan<T> implements PChanReceive<T>, PChanSend<T> {
        private promiseErrorTimeout;
        constructor(promiseErrorTimeout?: number);
        private closed;
        private onClosed;
        OnClosed: Promise<void>;
        private sentValues;
        private getValues;
        HasValues(): boolean;
        SendValue(...values: T extends void ? [] | [T | PromiseLike<T>] : [T | PromiseLike<T>]): void;
        SendError(error: any): void;
        GetPromise(): Promise<T>;
        IsClosed(): boolean;
        private closeSymbol;
        Close(): void;
        private errorOutOnClosed;
        IsClosedError(err: any): boolean;
    }
}
declare module "controlFlow/transform" {
    import { PChanReceive, PChanSend } from "controlFlow/pChan";
    export type TransformedChannel<Input, Output> = (input: Input) => Promise<Output>;
    /** Creates a channel that transforms the input, to the output, by calling the tranform function, but only in serial.
     *      I guess this is more like a real channel? Or it could be called a two way channel?
     */
    export function TransformChannel<Input, Output>(transform: (input: Input) => Promise<Output>, afterClose?: () => void): TransformedChannel<Input, Output> & {
        Close(): void;
    };
    export type TransformedChannelAsync<Input, Output> = (input: PChanReceive<Input>) => PChanReceive<Output>;
    /** Wraps a function (presumably an infinite loop) with logic to nicely bundle it up into something which transform an input
     *      channel into an output channel. Either channel being closed, or the main loop exiting closes everything.
     *
     * When the main promise resolves (or rejects), both channels will be closed, and nothing else should be attempted to be emitted.
     *
     * TODO: Add a while(true) transform mode, where we put some code in a while true loop, allowing errors to be caught,
     *      printed, and then the loop to be run again with new data.
     */
    export function TransformChannelAsync<Input, Output>(main: (params: {
        inputChan: PChanReceive<Input>;
        outputChan: PChanSend<Output>;
    }) => Promise<void>): TransformedChannelAsync<Input, Output>;
}
declare module "controlFlow/Deferred" {
    export class Deferred<T> {
        private value;
        private resolved;
        private resolveInternal;
        private rejectInternal;
        private promise;
        constructor();
        Resolve: (...values: T extends void ? [(T | PromiseLike<T> | undefined)?] : [T | PromiseLike<T>]) => this;
        Reject: (error: any) => this;
        /** Resolves it, using a new internal promise with a new value if this Deferred has previous been resolved or rejected. */
        ForceResolve: (...values: T extends void ? [(T | PromiseLike<T> | undefined)?] : [T | PromiseLike<T>]) => this;
        Promise: () => Promise<T>;
        Value: () => {
            v: T;
        } | {
            error: any;
        } | undefined;
    }
}
declare module "controlFlow/constructors" {
    import { PChan } from "controlFlow/pChan";
    export function PChanFromArray<T>(...values: T[]): PChan<T>;
    export function PChanFromGenerator<T>(generator: () => IterableIterator<T>): PChan<T>;
}
declare module "pchannel" {
    import { IsEmpty, Range } from "reflection/misc";
    import { pchan, PChan, SetDefaultTimeout } from "controlFlow/pChan";
    import { TransformChannel, TransformedChannel, TransformChannelAsync, TransformedChannelAsync } from "controlFlow/transform";
    import { Deferred } from "controlFlow/Deferred";
    import { PChanFromArray, PChanFromGenerator } from "controlFlow/constructors";
    const _ThrowIfNotImplementsData: (proposedData: Types.AnyAll, dataContract: Types.AnyAll) => void;
    const _AnyOrder: <T extends Types.AnyAll>(value: T) => T;
    const _ObjectExact: <T extends Types.AnyAll>(value: T) => T;
    const _Opt: <T extends Types.AnyAll>(value: T) => T;
    const _IsNumber: (str: string) => boolean;
    const _IsInteger: (num: number) => boolean;
    const _IsPrimitive: (value: Types.AnyAll) => value is Types.Primitive;
    const _CanHaveChildren: (value: Types.AnyAll) => value is Types.Dictionary;
    const _IsArray: (obj: Types.AnyAll) => obj is Types.Arr;
    const _SetTimeoutAsync: (time: number) => Promise<void>;
    const _Throws: (code: () => void) => void;
    const _ThrowsAsync: (code: () => Promise<void>) => Promise<void>;
    let _g: any;
    export { _ThrowIfNotImplementsData as ThrowIfNotImplementsData, _AnyOrder as AnyOrder, _ObjectExact as ObjectExact, _Opt as Opt, _IsNumber as IsNumber, _IsInteger as IsInteger, _IsPrimitive as IsPrimitive, _CanHaveChildren as CanHaveChildren, _IsArray as IsArray, _SetTimeoutAsync as SetTimeoutAsync, _Throws as Throws, _ThrowsAsync as ThrowsAsync, pchan, PChan, SetDefaultTimeout, IsEmpty, Range, TransformChannel, TransformedChannel, TransformChannelAsync, TransformedChannelAsync, Deferred, PChanFromArray, PChanFromGenerator, _g as g };
}


